C ----------------------------------------------------------------------
C
C Francais
C
C     ParaFit: Test de coevolution hote-parasite.
C
C     Ce programme test la signification des relations hote-parasite
C     individuelles entre un clade d'hotes et un clade de parasites.
C     
C H0: Le parasite s'attaque a des hotes choisis au hasard. En d'autres termes, 
C     les liens hotes-parasites ont des positions aleatoires. On obtiendra des 
C     realisations de H0 en permutant au hasard le contenu de chaque ligne de 
C     la matrice A. Les permutations sont effectuees de facon independante 
C     d'une ligne a l'autre.
C
C
C     Trois matrices de donnees sont lues par le programme:
C
C     A = matrice de valeurs {0, 1} decrivant les parasites qui se trouvent 
C         chez chaque hote. Cette matrice possede n1 lignes (n1 = nombre de 
C         parasites) et n2 colonnes (n2 = nombre d'hotes).
C     B = matrice des coordonnees principales, de taille (n1,n4), decrivant 
C         la phylogenie des parasites. Les coordonnees principales, calculees
C         a partir d'une matrice de distances phylogenetiques ou patristiques, 
C         sont corrigees pour les valeurs propres negatives si besoin est.
C     C = matrice transposee des coordonnees principales, de taille (n3,n2), 
C         decrivant la phylogenie des hotes. Les coordonnees principales, 
C         calculees a partir d'une matrice de distances phylogenetiques ou  
C         patristiques, sont corrigees pour les valeurs propres negatives 
C         si besoin est.
C     D = C A' B.  D est de taille (n3,n4).
C
C     Taille des matrices:   A(n1max,n2max)  B(n1max,n4max)
C                            C(n3max,n2max)  D(n3max,n4max)
C
C
C                                          Pierre Legendre, janvier 2002
C
C ----------------------------------------------------------------------
C
C English
C
C     ParaFit: A test of host-parasite coevolution.
C
C     This program computes tests of individual host-parasite relationships
C     involving a clade of hosts and a clade of parasites.
C
C H0: The parasites invade hosts chosen at random. In other words, the
C     host-parasite links have random positions. Realizations of H0 are
C     obtained by permuting at random the contents of each row of matrix A.
C     The permutations are done independently from row to row.
C
C
C     Three data matrices are read by the program:
C
C     A = matrix of {0, 1} values describing the parasites associated with 
C         each host. This matrix has n1 rows (n1 = number of parasites) 
C         and n2 columns (n2 = number of hosts).
C     B = Matrix of principal coordinates, of size (n1,n4), describing the
C         phylogeny of the parasites. The principal coordinates, computed
C         from a file of phylogenetic or patristic distances, are corrected
C         for negative eigenvalues if necessary.
C     C = transposed matrix of principal coordinates, of size (n3,n2), 
C         describing the  phylogeny of the hosts. The principal coordinates,
C         computed from a file of phylogenetic or patristic distances, are 
C         corrected for negative eigenvalues if necessary. 
C     D = C A' B.  D is of size (n3,n4).
C
C     Size of the matrices:   A(n1max,n2max)  B(n1max,n4max)
C                             C(n3max,n2max)  D(n3max,n4max)
C
C
C                                          Pierre Legendre, January 2002
C
C ----------------------------------------------------------------------
C
C Reference:
C
C Legendre, P., Y. Desdevises and E. Bazin. 2002. A statistical test for 
C    host-parasite coevolution. Systematic Biology 51(2): (in press).
C
C ----------------------------------------------------------------------
C2345678901234567890123456789012345678901234567890123456789012345678901234567890
      Parameter (n1max=512,n2max=512,n3max=n2max-1,n4max=n1max-1)
      Integer iordre(n2max),iparcnt(n1max),ihostcnt(n2max)
      Real*8 A(n1max,n2max),B(n1max,n4max),C(n3max,n2max),
     +       D(n3max,n4max),Aprime(n2max,n1max),CAprime(n3max,n1max),
     +       AprPer(n2max,n1max),eigvalB(n4max),eigvalC(n3max)
C
C Macintosh version: BBEdit is the 'creator' for output files
C      call F_Creator('R*ch')
C Langue du dialogue / Dialog language: 1 = francais; 2 = English
    1 write(*,*) 'Francais: tapez (1)'
      write(*,*) 'English:  type  (2)'
      read(*,*) langue
      if((langue.lt.1).or.(langue.gt.2)) goto 1
      write(*,*)
      if(langue.eq.1) then
         open(7,file='Hotepara.out',status='NEW')
         else
         open(7,file='Hostpara.out',status='NEW')
         endif
      if(langue.eq.1) then
         write(*,101)
         write(7,101)
         else
         write(*,201)
         write(7,201)
         endif
C Lecture des donnees / Read the data files
      call ReadData(A,B,C,n1max,n2max,n3max,n4max,n1,n2,n3,n4,langue)
C      write(*,*)
C      write(*,*) 'Informations additionnelles (debug)? (0) non, (1) oui'
C      read(*,*) idebug
      idebug=0
      write(*,*)
      if(langue.eq.1) write(*,*) 'Combien de permutations? ',
     +   '(e.g., 999, 9999, ...)'
      if(langue.eq.2) write(*,*) 'How many permutations? ',
     +   '(e.g., 999, 9999, ...)'
      read(*,*) nperm
      if(nperm.lt.0) then
         idebug=1
         nperm=-nperm
         endif
      if(langue.eq.1) then
         write(7,108) nperm
         write(*,109)
         else
         write(7,208) nperm
         write(*,209)
         endif
C Demarrer l'horloge / Start the clock
      debut=secnds(0.0)
      call Compute(A,B,C,D,Aprime,CAprime,AprPer,n1max,n2max,n3max,
     +   n4max,n1,n2,n3,n4,iordre,iparcnt,ihostcnt,nperm,langue,idebug,
     +   eigvalB,eigvalC)
C Fin du calcul; lire l'horloge / End of computations; read the clock
      write(*,*)
      fin=secnds(debut)
      if(langue.eq.1) then
         write(*,110) fin
         write(7,*)
         write(7,110) fin
         write(*,*) 
     +   "Les resultats se trouvent dans le fichier 'Hotepara.out'"
         write(*,*) 'Fin du programme.'
         else
         write(*,210) fin
         write(7,*)
         write(7,210) fin
         write(*,*) "Results are found in file 'Hostpara.out'"
         write(*,*) 'End of the program.'
         endif
      close(7)
C
      stop
  101 format(' ParaFit: Test de coevolution hote-parasite'//
     +       ' Pierre Legendre'/
     +       ' Departement de sciences biologiques'/
     +       ' Universite de Montreal.'/
     +       '             (c) Pierre Legendre, 2001'/)
  201 format(' ParaFit: A test of host-parasite coevolution'//
     +       ' Pierre Legendre'/
     +       ' Departement de sciences biologiques'/
     +       ' Universite de Montreal.'/
     +       '             (c) Pierre Legendre, 2001'/)
  108 format(' Nombre de permutations:',i6/)
  208 format(' Number of permutations:',i6/)
  109 format(/' Un moment SVP...'/)
  209 format(/' Please wait...'/)
  110 format(' Duree du calcul:',f10.2,' sec.'/)
  210 format(' Computation time:',f10.2,' sec.'/)
      end
      
      Subroutine Compute(A,B,C,D,Aprime,CAprime,AprPer,n1max,n2max,
     +   n3max,n4max,n1,n2,n3,n4,iordre,iparcnt,ihostcnt,nperm,
     +   langue,idebug,eigvalB,eigvalC)
      Parameter (maxperm=10000)
      Integer iordre(n2max),iparcnt(n1max),ihostcnt(n2max)
      Real*8 A(n1max,n2max),B(n1max,n4max),C(n3max,n2max),
     +       D(n3max,n4max),Aprime(n2max,n1max),CAprime(n3max,n1max),
     +       AprPer(n2max,n1max)
      Real*8 TraceD,trace0,trac0per,tracetot,traceper(maxperm),temp,
     +       F1,F2,F1per,F2per,prob,prob1,prob2,eigvalB(n4max),
     +       eigvalC(n3max),ssqeigvB,ssqeigvC
      Character nom*10
C
C This subroutine computes
C the general ParaFit test as well as the tests for individual links.
C The 'traceper' values needed to calculate F1 and F2 are computed only once.
      if(nperm+1.gt.maxperm) then
         write(*,*) 'Too many permutations. Increase the length of'
         write(*,*) 'vector "traceper" and recompile the program.'
         endif
C Initialisation des permutations / Initialize seed for permutations
C Initialisation de 'ran2' en donnant a 'idum' la valeur d'un entier negatif
      idum=-1
      nturn=10
      do 1 i=1,nturn
    1 turn=ran2(idum)
C tracemax = max(sum of squared PCoA eigenvalues of B, 
C                sum of squared PCoA eigenvalues of C)
      do 2 j=1,n4
    2 eigvalB(j)=0.0
      do 4 i=1,n1
      do 4 j=1,n4
    4 eigvalB(j)=eigvalB(j)+B(i,j)*B(i,j)
      if(idebug.eq.1) write(7,112) (eigvalB(j), j=1,n4)
  112 format(' PCoA eigenvalues of B:'/5(10f10.5))
      ssqeigvB=0.0
      do 6 j=1,n4
    6 ssqeigvB=ssqeigvB+eigvalB(j)*eigvalB(j)
      if(idebug.eq.1) write(7,107) ssqeigvB
  107 format(' Sum of squared PCoA eigenvalues of B =',f15.5/)
C
      do 8 i=1,n3
    8 eigvalC(i)=0.0
      do 10 i=1,n3
      do 10 j=1,n2
   10 eigvalC(i)=eigvalC(i)+C(i,j)*C(i,j)
      if(idebug.eq.1) write(7,108) (eigvalC(i), i=1,n3)
  108 format(' PCoA eigenvalues of C:'/5(10f10.5))
      ssqeigvC=0.0
      do 12 i=1,n3
   12 ssqeigvC=ssqeigvC+eigvalC(i)*eigvalC(i)
      if(idebug.eq.1) write(7,107) ssqeigvC
  109 format(' Sum of squared PCoA eigenvalues of C =',f15.5/)
C
      tracetot=max(ssqeigvB,ssqeigvC)
      if(idebug.eq.1) write(7,110) tracetot
  110 format(' TraceTot =',f15.5/)

C Global test. Statistic: ParaFitGlobal = trace(D).
C Compute the pseudo-trace of D for the original matrix A
C     D = C A' B
      do 14 i=1,n1
      do 14 j=1,n2
   14 Aprime(j,i)=A(i,j)
      call PRODUCT(C,Aprime,CAprime,n3,n2,n1,n3max,n2max,n1max)
      call PRODUCT(CAprime,B,D,n3,n1,n4,n3max,n1max,n4max)
      traceper(1)=TraceD(D,n3,n4,n3max,n4max)
C Compute trace of D for the permuted matrix A, and test the trace of D
      iGE=1
      do 16 i=1,n2
   16 iordre(i)=i
C Begin permutations - - - - - - - - - - -
      do 30 iperm=1,nperm
      do 20 i=1,n1
      call Permute(idum,n2,n2max,iordre)
      do 18 j=1,n2
   18 AprPer(j,i)=Aprime(iordre(j),i)
   20 continue
      call PRODUCT(C,AprPer,CAprime,n3,n2,n1,n3max,n2max,n1max)
      call PRODUCT(CAprime,B,D,n3,n1,n4,n3max,n1max,n4max)
      temp=TraceD(D,n3,n4,n3max,n4max)
      traceper(iperm+1)=temp
      if(temp.ge.traceper(1)) iGE=iGE+1
   30 continue
C End permutations - - - - - - - - - - - -
      prob=dfloat(iGE)/dfloat(nperm+1)
      if(langue.eq.1) then
         write(7,105) traceper(1),prob
         write(7,106)
         else
         write(7,205) traceper(1),prob
         write(7,206)
         endif

C Test of individual H-P associations:
C Remove each host-parasite link in turn.
C F1 = ParaFitLink1, F2 = ParaFitLink2.
      icount=0
C
      do 50 ii=1,n1
      do 50 jj=1,n2
      if(A(ii,jj).eq.0.0) goto 50
      icount=icount+1
      do 32 i=1,n1
      do 32 j=1,n2
   32 Aprime(j,i)=A(i,j)
      Aprime(jj,ii)=0.0
      call PRODUCT(C,Aprime,CAprime,n3,n2,n1,n3max,n2max,n1max)
      call PRODUCT(CAprime,B,D,n3,n1,n4,n3max,n1max,n4max)
      trace0=TraceD(D,n3,n4,n3max,n4max)
      F1=(traceper(1)-trace0)
      F2=(traceper(1)-trace0)/(tracetot-traceper(1))
      if(idebug.eq.1) write(7,111) traceper(1),trace0
  111 format(/'   traceper(1) =',f10.5,'   trace0 =',f10.5)
C Compute trace of permuted matrix A with 1 link removed, and test it
      iGEF1=1
      iGEF2=1
C Reinitialisation des permutations / Reinitialize seed for permutations
C Initialisation de 'ran2' en donnant a 'idum' la valeur d'un entier negatif
C
C Cela permet d'obtenir la meme sequence de nombres aleatoires lors des tests
C des liens H-P particuliers que celle qui a t utilisee lors du test
C global, ce qui autorise a reutiliser pour le calcul de F1 et F2 les 
C valeurs du vecteur 'traceper' calculees plus haut.
      idum=-1
      do 33 i=1,nturn
   33 turn=ran2(idum)
      do 34 i=1,n2
   34 iordre(i)=i
C Begin permutations - - - - - - - - - - -
      do 40 iperm=1,nperm
      do 38 i=1,n1
      call Permute(idum,n2,n2max,iordre)
      do 36 j=1,n2
   36 AprPer(j,i)=Aprime(iordre(j),i)
   38 continue
      call PRODUCT(C,AprPer,CAprime,n3,n2,n1,n3max,n2max,n1max)
      call PRODUCT(CAprime,B,D,n3,n1,n4,n3max,n1max,n4max)
      trac0per=TraceD(D,n3,n4,n3max,n4max)
      F1per=(traceper(iperm+1)-trac0per)
      F2per=(traceper(iperm+1)-trac0per)/(Tracetot-traceper(iperm+1))
C      F2per=(traceper(iperm+1)-trac0per)/traceper(iperm+1)
      if(F1per.ge.F1) iGEF1=iGEF1+1
      if(F2per.ge.F2) iGEF2=iGEF2+1
   40 continue
C End permutations - - - - - - - - - - - -
      prob1=dfloat(iGEF1)/dfloat(nperm+1)
      prob2=dfloat(iGEF2)/dfloat(nperm+1)
      if(langue.eq.1) then
         write(7,101) ii,jj,F1,prob1,F2,prob2
         else
         write(7,201) ii,jj,F1,prob1,F2,prob2
         endif
   50 continue

C Print additional statistical information
      do 52 i=1,n1
   52 iparcnt(i)=0
      do 54 j=1,n2
   54 ihostcnt(j)=0
      do 58 i=1,n1
      do 58 j=1,n2
      iparcnt (i)=iparcnt (i)+A(i,j)
   58 ihostcnt(j)=ihostcnt(j)+A(i,j)
      if(langue.eq.1) then
         write(7,102) icount
         write(7,103) (iparcnt (i), i=1,n1)
         write(7,104) (ihostcnt(j), j=1,n2)
         else
         write(7,202) icount
         write(7,203) (iparcnt (i), i=1,n1)
         write(7,204) (ihostcnt(j), j=1,n2)
         endif
C
      return
  101 format(' Parasite',i3,'  Hote',i3,'   F1 =',
     +   f10.5,'   Prob1 =',f8.5,'   F2 =',f10.5,'   Prob2 =',f8.5)
  102 format(/' Il y a',i4,' liens hote-parasite dans la matrice A')
  103 format(/' Nombre d''hotes par parasite:'/50i3)
  104 format(/' Nombre de parasites par hote:'/50i3)
  105 format(' Test global de cospeciation:                    ',
     +       ' ParaFitGlobal =',f8.5,'   Prob  =',f8.5/)
  106 format(' Test des liens hote-parasite individuels:'//
     +       '                        F1 = ParaFitLink1',
     +       '                  F2 = ParaFitLink2'/)
  201 format(' Parasite',i3,'  Host',i3,'   F1 =',
     +   f10.5,'   Prob1 =',f8.5,'   F2 =',f10.5,'   Prob2 =',f8.5)
  202 format(/' There are',i4,' host-parasite links in matrix A')
  203 format(/' Number of hosts per parasite:'/50i3)
  204 format(/' Number of parasites per host:'/50i3)
  205 format(' Global test of cospeciation:                    ',
     +       ' ParaFitGlobal =',f8.5,'   Prob  =',f8.5/)
  206 format(' Test of individual host-parasite links:'//
     +       '                        F1 = ParaFitLink1',
     +       '                  F2 = ParaFitLink2'/)
      end
 
      Subroutine Permute(idum,n,nmax,iordre)
C This subroutine permutes the first 'n' values of vector 'iordre' at random
C in an equiprobable way. This property has been checked through intensive 
C simulations.
      Integer idum,n,i,itemp,km1,m
      Integer iordre(nmax)
      m=n
      km1=n-1
      do 10 i=1,km1
    8    j = 1 + ran2(idum)*m
         if(j.gt.m) goto 8
         itemp = iordre(m)
         iordre(m) = iordre(j)
         iordre(j) = itemp
         m=m-1
   10 continue
      return
      end

      Subroutine Print(A,n1,n2,n1max,n2max,nom)
C Prints matrices. For debugging purpose.
      Real*8 A(n1max,n2max)
      Character nom*10
      write(8,100) nom
      do 10 i=1,n1
      write(8,101) (A(i,j), j=1,n2)
   10 continue
      write(8,*)
      return
  100 format(1x,a10)
  101 format(18f8.3)
      end
      
      Subroutine PRODUCT(Z1,Z2,Z3,n1,n2,n3,n1max,n2max,n3max)
      Real*8 Z1(n1max,n2max),Z2(n2max,n3max),Z3(n1max,n3max),temp
C Z1 * Z2 = Z3
      do 8 i=1,n1
      do 8 j=1,n3
      temp=0.0
      do 6 k=1,n2
    6 temp=temp+Z1(i,k)*Z2(k,j)
    8 Z3(i,j)=temp
      return
      end

C Subroutine Ran2 comes from Numerical Recipes.
C
C The author of the present program is not authorized to distribute the source
C code of Numerical Recipes subroutines. If you want to use this subroutine to
C recompile the present program (in the original or some modified form) or in 
C any other program, you should obtain a license from Numerical Recipes. 
C See the WWWeb site <http://world.std.com/~nr/storefront.html>.
C This Numerical Recipes subroutine is reproduced here to show how it is 
C used in the present program.
C
      Function Ran2(idum)
C Long period (> 2 x 10^18) random number generator of L'Ecuyer with
C Bays-Durham shuffle and added safeguards. Returns a uniform random
C deviate between 0.0 and 1.0 (exclusive of endpoint values). Call
C with idum a negative integer to initialize; thereafter, do not alter
C idum between successive deviates in a sequence. RNMX should approximate
C the largest floating value that is less than 1. Note: this was obtained
C from p. 272 of "Numerical Recipes in FORTRAN", 2nd ed. 1992.
      Integer idum,IM1,IM2,IMM1,IA1,IA2,IQ1,IQ2,IR1,IR2,NTAB,NDIV
      Real ran2,AM,EPS,RNMX
      Parameter (IM1=2147483563,IM2=2147483399, AM=1./IM1,IMM1=IM1-1,
     +    IA1=40014,IA2=40692,IQ1=53668,IQ2=52774,IR1=12211,
     +    IR2=3791,NTAB=32,NDIV=1+IMM1/NTAB,EPS=1.2E-7,RNMX=1.-EPS)
      Integer idum2,j,k,iv(NTAB),iy
      Save iv,iy,idum2
C These Data are a safety. Used for initialization only when the first idum >= 0.
      Data idum2/123456789/,iv/NTAB*0/,iy/0/
C
      if (idum.le.0) then
         idum=max(-idum,1)
         idum2=idum
         do j=NTAB+8,1,-1
            k=idum/IQ1
            idum=IA1*(idum-k*IQ1)-k*IR1
            if (idum.lt.0) idum=idum+IM1
            if (j.le.NTAB) iv(j)=idum
            enddo
         iy=iv(1)
         end if
C
      k=idum/IQ1
      idum=IA1*(idum-k*IQ1)-k*IR1
      if (idum.lt.0) idum=idum+IM1
      k=idum2/IQ2
      idum2=IA2*(idum2-k*IQ2)-k*IR2
      if (idum2.lt.0) idum2=idum2+IM2
      j=1+iy/NDIV
      iy=iv(j)-idum2
      iv(j)=idum
      if (iy.lt.1) iy=iy+IMM1
      ran2=min(AM*iy,RNMX)
C
      return
      end

      Subroutine ReadData(A,B,C,n1max,n2max,n3max,n4max,n1,n2,n3,n4,
     +           langue)
      Real*8 A(n1max,n2max),B(n1max,n4max),C(n3max,n2max)
      Character nom*10
      Character*79 namea
C Size of the matrices?  
    4 if(langue.eq.1) then
         write(*,*)
         write(*,*) 'Tapez n1, n2, n3, n4.'
         else
         write(*,*) 'Type n1, n2, n3, n4.'
         endif
      write(*,*)'   A(n1,n2)  B(n1,n4)'
      write(*,*)'   C(n3,n2)  D(n3,n4)'
      read(*,*) n1,n2,n3,n4
      write(*,*)
      if((n1.gt.n1max).or.(n2.gt.n2max).or.
     +   (n3.gt.n3max).or.(n4.gt.n4max))then
         if(langue.eq.1) then
            stop 'Matrice trop grande. Redimensionnez le programme.'
            else
            stop 'Matrix too large. Redimension the program.'
            endif
         endif
      if((n1.lt.0).or.(n2.lt.0).or.(n3.lt.0).or.(n4.lt.0)) then 
         if(langue.eq.1) then
            write(*,*) 'Mauvaise reponse: dimension <= 0. Reprenons...'
            else
            write(*,*) 'Wrong answer: dimension <= 0. Try again...'
            endif
         goto 4
         endif
C Read in the data tables --
C Matrix A:
      if(langue.eq.1) then
         write(*,*) 'Nom du fichier de donnees A(n1,n2) ?'
         else
         write(*,*) 'Name of input data file A(n1,n2) ?'
         endif
      read(*,*) namea
      open(1,File=namea,Status='OLD')
      inquire(unit=1,name=namea)
      write(*,*) namea
      if(langue.eq.1) then
         write(7,103) 1hA,namea
         else
         write(7,203) 1hA,namea
         endif
      do 12 i=1,n1
   12 read(1,*) (A(i,j), j=1,n2)
C      nom=' A =      '
C      call Print(A,n1,n2,n1max,n2max,nom)
C Matrix B:
      if(langue.eq.1) then
         write(*,*) 'Nom du fichier de donnees B(n1,n4) ?'
         else
         write(*,*) 'Name of input data file B(n1,n4) ?'
         endif
      read(*,*) namea
      open(2,File=namea,Status='OLD')
      inquire(unit=2,name=namea)
      write(*,*) namea
      if(langue.eq.1) then
         write(7,103) 1hB,namea
         else
         write(7,203) 1hB,namea
         endif
      do 14 i=1,n1
   14 read(2,*) (B(i,j), j=1,n4)
C      nom=' B =      '
C      call Print(B,n1,n4,n1max,n4max,nom)
C Matrix C:
      if(langue.eq.1) then
         write(*,*) 'Nom du fichier de donnees C(n3,n2) ?'
         else
         write(*,*) 'Name of input data file C(n3,n2) ?'
         endif
      read(*,*) namea
      open(3,File=namea,Status='OLD')
      inquire(unit=3,name=namea)
      write(*,*) namea
      if(langue.eq.1) then
         write(7,103) 1hC,namea
         else
         write(7,203) 1hC,namea
         endif
      write(7,*)
      do 18 i=1,n3
   18 read(3,*) (C(i,j), j=1,n2)
C      nom=' C =      '
C      call Print(C,n3,n2,n3max,n2max,nom)
      if(langue.eq.1) then
         write(*,*)'Fin de la lecture des matrices.'
         else
         write(*,*)'The input data matrices have been read.'
         endif
      close(1)
      close(2)
      close(3)
      return
  103 format(' Matrice ',A1,' = ',A32)
  203 format(' Matrix ',A1,' = ',A32)
      end

      Function TraceD(D,n3,n4,n3max,n4max)
      Real*8 TraceD,D(n3max,n4max),temp
C Trace statistic for a matrix D in which the column vectors are centred 
C on mean, or pseudo-trace using non-centred values (as used in ParaFit).
      temp=0.0
      do 10 i=1,n3
      do 10 j=1,n4
   10 temp=temp+D(i,j)*D(i,j)
      TraceD=temp
      return
      end
